// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"encoding/json"
	"fmt"
	"io"
	"os"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
)

var autodocCmd = cobra.Command{
	Use:   "autodoc",
	Short: "Autodocument current configuration",
	Long:  "Extract the current configuration (actors, pipelines etc) for documentation purpose",
	Run: func(_ *cobra.Command, _ []string) {
		format := viper.GetString("format")
		if format == "" {
			format = "json"
		}
		output := viper.GetString("output")
		if output == "" {
			if format == "json" {
				output = "-"
			} else {
				output = "./doc"
			}
		}
		os.Exit(autodoc(output, format))
	},
}

type Documentation struct {
	ServerURL    string              `json:"server-url"`
	AccountList  []*api.Account      `json:"accounts"`
	ActorList    []*api.Actor        `json:"actors"`
	PipelineList []*storage.Pipeline `json:"pipelines"`
}

func fetchDocumentation(conn *xbus.Client) (*Documentation, error) {
	Logger.Info().Msg("Getting accounts...")
	accounts, err := conn.AccountList()
	if err != nil {
		return nil, fmt.Errorf("could not retrieve accounts: %w", err)
	}

	Logger.Info().Msg("Getting actors...")
	actors, err := conn.ActorList([]*api.Actor{})
	if err != nil {
		return nil, fmt.Errorf("could not retrieve actors: %w", err)
	}

	Logger.Info().Msg("Getting pipelines...")
	pipelines, err := conn.PipelineQuery("")
	if err != nil {
		return nil, fmt.Errorf("could not retrieve pipelines: %w", err)
	}

	doc := Documentation{
		ServerURL:    conn.GetOptions().ConnOptions.Url,
		ActorList:    actors,
		AccountList:  accounts,
		PipelineList: make([]*storage.Pipeline, len(pipelines)),
	}
	for i, info := range pipelines {
		graph, err := pipelineGetGraph(conn, info)
		if err != nil {
			return nil, err
		}
		doc.PipelineList[i] = &storage.Pipeline{
			PipelineInfo: info,
			Graph:        *graph,
		}
	}

	return &doc, nil
}

func openOutputFile(name string) (io.WriteCloser, error) {
	if name == "-" {
		return os.Stdout, nil
	}
	f, err := os.Create(name)
	if err != nil {
		return nil, err
	}
	return f, nil
}

func autodoc(output, format string) int {
	if format != "json" {
		Logger.Error().Msgf("invalid format '%s'. Expected 'json'.", format)
		return 1
	}

	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 2
	}

	doc, err := fetchDocumentation(conn)
	if err != nil {
		Logger.Err(err).Msg("Could not retrieve the documentation")
		return 2
	}

	switch format {
	case "json":
		f, err := openOutputFile(output)
		if err != nil {
			Logger.Err(err).Msg("Could not open output file")
		}
		defer func() {
			if err := f.Close(); err != nil {
				Logger.Err(err).Msg("Could not close output file")
			}
		}()

		encoder := json.NewEncoder(f)
		encoder.SetIndent("", "  ")
		if err := encoder.Encode(doc); err != nil {
			Logger.Err(err).Msg("Could not marshal the documentation")
		}
	default:
		panic("unknown documentation format")
	}

	return 0
}

func init() {
	RootCmd.AddCommand(&autodocCmd)

	flags := autodocCmd.PersistentFlags()
	flags.StringP("output", "o", "-", "Output file or directory")
	must(viper.BindPFlag("output", flags.Lookup("output")))
	flags.StringP("format", "f", "json", "Output format")
	must(viper.BindPFlag("format", flags.Lookup("format")))
}
