// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package http

import (
	"bytes"
	"context"
	"errors"
	"fmt"
	"io"
	"net/http"
	"strconv"
	"strings"
	"time"

	nrpc "github.com/nats-rpc/nrpc"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"google.golang.org/protobuf/proto"
	"google.golang.org/protobuf/types/known/timestamppb"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/envelope"
)

var (
	ErrStopIteration = errors.New("stop iteration")
	ErrNoSuchInput   = errors.New("no such input")
)

type WebHookPayloadType string

const (
	// WebHookServiceName is the name of the webhook service in the registry
	WebHookServiceName = "http.webhook"

	WebHookPayloadFullRequest       WebHookPayloadType = "full-request"
	WebHookPayloadEnvelope          WebHookPayloadType = "envelope"
	WebHookPayloadMessageInRequest  WebHookPayloadType = "message-in-request"
	WebHookPayloadMessageInEnvelope WebHookPayloadType = "message-in-envelope"
	WebHookPayloadMessageContent    WebHookPayloadType = "message-content"
)

var ErrInvalidWebHookPayloadType = errors.New("invalid webhook payload type")

func WebHookPayloadTypeFromString(s string) (WebHookPayloadType, error) {
	switch t := WebHookPayloadType(s); t {
	case
		WebHookPayloadFullRequest,
		WebHookPayloadEnvelope,
		WebHookPayloadMessageInRequest,
		WebHookPayloadMessageInEnvelope,
		WebHookPayloadMessageContent:
		return t, nil
	case "":
		return WebHookPayloadFullRequest, nil
	default:
		return WebHookPayloadType(""), fmt.Errorf("%w: %s", ErrInvalidWebHookPayloadType, s)
	}
}

type Header struct {
	Key   string
	Value string
}

// WebHookHandler ...
type WebHookHandler struct {
	url               string
	headers           []Header
	contentType       string
	jsonItemEncodings []api.ItemEncoding
	requestMarshaler  func(r *api.ActorProcessRequest) ([]byte, error)
	envelopeMarshaler func(context.Context, envelope.Envelope) (io.ReadCloser, error)

	payloadType WebHookPayloadType

	successHTTPCodes          []int
	temporaryFailureHTTPCodes []int

	retryInitialDelay time.Duration
	retryMaxDelay     time.Duration
	retryTimeout      time.Duration
	retryBackoff      int
}

// NewWebHookHandler ...
func NewWebHookHandler(actor *xbus.Actor, _ xbus.Settings) (xbus.ActorService, error) {
	url := actor.Options.GetConfigEntry("http.webhook.url")
	if url == "" {
		actor.Log.Error().Msgf("No webhook on this actor")
		return nil, fmt.Errorf("no webhook on this actor")
	}

	headers, err := parseHeadersList(
		actor.Options.GetConfigEntry("http.webhook.headers"))
	if err != nil {
		return nil, fmt.Errorf("error parsing http.webhook.headers: %w", err)
	}

	contentType := actor.Options.GetConfigEntry("http.webhook.content-type")
	switch contentType {
	case "", "json":
		contentType = "application/json"
	case "proto", "protobuf":
		contentType = "application/x-protobuf"
	}

	jsonItemEncodingEntry := actor.Options.GetConfigEntry("http.webhook.json-item-encoding")
	if jsonItemEncodingEntry == "" {
		jsonItemEncodingEntry = "binary"
	}
	jsonItemEncodings, err := api.ParseItemEncodingList(jsonItemEncodingEntry)
	if err != nil {
		return nil, fmt.Errorf("error parsing http.weebhook.json-item-encoding")
	}

	successHTTPCodes, err := parseHTTPCodesList(
		actor.Options.GetConfigEntry("http.webhook.http-codes.success"))
	if err != nil {
		return nil, fmt.Errorf("error parsing http.webhook.http-codes.success: %w", err)
	}

	temporaryFailureHTTPCodes, err := parseHTTPCodesList(
		actor.Options.GetConfigEntry("http.webhook.http-codes.temporary-failure"))
	if err != nil {
		return nil, fmt.Errorf("error parsing http.webhook.http-codes.temporary-failure: %w", err)
	}

	retryBackoff, err := parsePercentage(
		actor.Options.GetConfigEntry("http.webhook.retry.backoff"),
		50)
	if err != nil {
		return nil, fmt.Errorf("error parsing http.webhook.retry.backoff: %w", err)
	}
	retryInitialDelay, err := parseDuration(
		actor.Options.GetConfigEntry("http.webhook.retry.initial-delay"),
		time.Second)
	if err != nil {
		return nil, fmt.Errorf("error parsing http.webhook.retry.initial-delay: %w", err)
	}
	retryMaxDelay, err := parseDuration(
		actor.Options.GetConfigEntry("http.webhook.retry.max-delay"),
		5*time.Minute)
	if err != nil {
		return nil, fmt.Errorf("error parsing http.webhook.retry.max-delay: %w", err)
	}
	retryTimeout, err := parseDuration(
		actor.Options.GetConfigEntry("http.webhook.retry.timeout"),
		time.Hour)
	if err != nil {
		return nil, fmt.Errorf("error parsing http.webhook.retry.timeout: %w", err)
	}

	payloadType, err := WebHookPayloadTypeFromString(
		actor.Options.GetConfigEntry("http.webhook.payload-type"))
	if err != nil {
		return nil, fmt.Errorf("error parsing http.webhook.payload-type: %w", err)
	}

	if len(successHTTPCodes) == 0 {
		switch payloadType {
		case WebHookPayloadFullRequest, WebHookPayloadMessageInRequest:
			successHTTPCodes = []int{200}
		case WebHookPayloadEnvelope, WebHookPayloadMessageInEnvelope, WebHookPayloadMessageContent:
			successHTTPCodes = []int{204}
		}
	}

	handler := WebHookHandler{
		url:                       url,
		headers:                   headers,
		contentType:               contentType,
		jsonItemEncodings:         jsonItemEncodings,
		requestMarshaler:          requestMarshaler(contentType, jsonItemEncodings),
		envelopeMarshaler:         envelopeMarshaler(contentType, jsonItemEncodings),
		successHTTPCodes:          successHTTPCodes,
		temporaryFailureHTTPCodes: temporaryFailureHTTPCodes,
		payloadType:               payloadType,
		retryBackoff:              retryBackoff,
		retryInitialDelay:         retryInitialDelay,
		retryMaxDelay:             retryMaxDelay,
		retryTimeout:              retryTimeout,
	}
	return &handler, nil
}

func requestMarshaler(contentType string, jsonItemEncodings []api.ItemEncoding) func(r *api.ActorProcessRequest) ([]byte, error) {
	switch contentType {
	case "application/json":
		return jsonMarshalRequest(jsonItemEncodings)
	case "application/x-protobuf":
		return protobufMarshalRequest
	default:
		panic("no marshaler for contenttype: " + contentType)
	}
}

func jsonMarshalRequest(jsonItemEncodings []api.ItemEncoding) func(r *api.ActorProcessRequest) ([]byte, error) {
	return func(r *api.ActorProcessRequest) ([]byte, error) {
		bytes, err := api.JSONMarshal(jsonItemEncodings, r)
		if err != nil {
			return nil, fmt.Errorf("[http.webhook] error encoding the payload: %w", err)
		}
		return bytes, nil
	}
}

func protobufMarshalRequest(r *api.ActorProcessRequest) ([]byte, error) {
	bytes, err := proto.Marshal(r)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] error encoding the payload: %w", err)
	}
	return bytes, nil
}

func envelopeMarshaler(contentType string, itemEncoding []api.ItemEncoding) func(context.Context, envelope.Envelope) (io.ReadCloser, error) {
	switch contentType {
	case "application/json":
		return jsonMarshalEnvelope(itemEncoding)
	case "application/x-protobuf":
		return protobufMarshalEnvelope
	default:
		panic("no marshaler for contenttype: " + contentType)
	}
}

func jsonMarshalEnvelope(itemEncoding []api.ItemEncoding) func(context.Context, envelope.Envelope) (io.ReadCloser, error) {
	return func(ctx context.Context, e envelope.Envelope) (io.ReadCloser, error) {
		complete, err := envelope.CompleteWithContext(ctx, e)
		if err != nil {
			return nil, fmt.Errorf("[http.webhook] failed to read the envelope: %w", err)
		}
		bytes, err := api.JSONMarshal(itemEncoding, complete)
		if err != nil {
			return nil, fmt.Errorf("[http.webhook] error encoding the payload: %w", err)
		}
		r, w := io.Pipe()
		go func() {
			// TODO implement as a stream
			if _, err := w.Write(bytes); err != nil {
				// the error should also be returned to the reader, which should
				// report it. We can ignore it.
				zerolog.Ctx(ctx).Err(err).Msg("[http.webhook] could not write the bytes down the pipe")
			}
			if err := w.Close(); err != nil {
				// the error should also be returned to the reader, which should
				// report it. We can ignore it.
				zerolog.Ctx(ctx).Err(err).Msg("[http.webhook] could not close the pipe")
			}
		}()
		return r, nil
	}
}

func protobufMarshalEnvelope(ctx context.Context, e envelope.Envelope) (io.ReadCloser, error) {
	complete, err := envelope.CompleteWithContext(ctx, e)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] failed to read the envelope: %w", err)
	}
	bytes, err := proto.Marshal(complete)
	if err != nil {
		return nil, fmt.Errorf("[http.webhook] error encoding the payload: %w", err)
	}
	r, w := io.Pipe()
	go func() {
		if _, err := w.Write(bytes); err != nil {
			// the error should also be returned to the reader, which should
			// report it. We can ignore it.
			zerolog.Ctx(ctx).Err(err).Msg("[http.webhook] could not write the bytes down the pipe")
		}
		if err := w.Close(); err != nil {
			// the error should also be returned to the reader, which should
			// report it. We can ignore it.
			zerolog.Ctx(ctx).Err(err).Msg("[http.webhook] could not close the pipe")
		}
	}()
	return r, nil
}

func unmarshalActorProcessingState(contentType string, body io.Reader) (*api.ActorProcessingState, error) {
	var state api.ActorProcessingState
	switch contentType {
	case "application/json":
		if err := api.DecodeJSON(body, &state); err != nil {
			return nil, fmt.Errorf("[http.webhook] error decoding json response: %w", err)
		}
	case "application/x-protobuf":
		b, err := io.ReadAll(body)
		if err != nil {
			return nil, fmt.Errorf("[http.webhook] error reading response: %w", err)
		}
		if err := proto.Unmarshal(b, &state); err != nil {
			return nil, fmt.Errorf("[http.webhook] error decoding protobuf response: %w", err)
		}
	case "":
		return nil, fmt.Errorf("[http.webhook] missing Content-Type header in the response from the webhook")
	default:
		return nil, fmt.Errorf("[http.webhook] invalid Content-Type header in the response from the webhook. Expected 'application/json' or 'application/x-protobuf', got %s", contentType)
	}
	return &state, nil
}

func (wh *WebHookHandler) postRequest(
	ctx context.Context, msgtype string, r *api.ActorProcessRequest, log zerolog.Logger,
) (*api.ActorProcessingState, error) {
	var state *api.ActorProcessingState

	url := strings.ReplaceAll(wh.url, "{msgtype}", msgtype)

	payload, err := wh.requestMarshaler(r)
	if err != nil {
		return state, err
	}

	req, err := http.NewRequest("POST", url, bytes.NewBuffer(payload))
	if err != nil {
		return state, fmt.Errorf("failed to forge the request: %w", err)
	}
	req = req.WithContext(ctx)
	req.Header.Set("Content-Type", wh.contentType)
	req.Header.Set("Accept", wh.contentType)
	for _, h := range wh.headers {
		req.Header.Set(h.Key, h.Value)
	}

	if log.GetLevel() <= zerolog.DebugLevel {
		inputs := make([]string, len(r.GetInputs()))
		for i, input := range r.GetInputs() {
			inputs[i] = input.GetName() + "/" + input.Envelope.GetIDAsUUID().String()
		}
		log.Debug().
			Str("url", url).
			Strs("inputs", inputs).
			Str("content-type", wh.contentType).
			Msg("posting request")
	}
	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return state, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] %s", err.Error()),
		}
	}
	defer resp.Body.Close()
	if httpCodeMatch(resp.StatusCode, wh.temporaryFailureHTTPCodes) {
		var (
			message = resp.Status
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message += string(buf[:n])
		} else {
			message += ("Could not read body: " + err.Error())
		}
		return state, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] Temporary failure: %s", message),
		}
	}

	if httpCodeMatch(resp.StatusCode, wh.successHTTPCodes) {
		contentType := resp.Header.Get("Content-Type")
		state, err = unmarshalActorProcessingState(contentType, resp.Body)
		if err != nil {
			state = &api.ActorProcessingState{
				Status: api.ActorProcessingState_ERROR,
				Messages: []*api.LogMessage{{
					Time:  timestamppb.New(time.Now()),
					Level: api.LogLevel_ERROR,
					Text:  "[http.webhook] " + err.Error(),
				}},
			}
			return state, err
		} else if state.Status == api.ActorProcessingState_ERROR && len(state.Messages) == 0 {
			state.Messages = []*api.LogMessage{
				{
					Time:  timestamppb.Now(),
					Level: api.LogLevel_ERROR,
					Text:  "[http.webhook] The webhook returned a 'ERROR' state without any error message",
				},
			}
		}
	} else {
		var (
			message string
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message = string(buf[:n])
		} else {
			message = "Could not read body: " + err.Error()
		}
		message = fmt.Sprintf(`[http.webhook] Invalid http status. Expected one of %v, got %s.
The response body is:
%s`,
			wh.successHTTPCodes,
			resp.Status,
			message)
		state = &api.ActorProcessingState{
			Status: api.ActorProcessingState_ERROR,
			Messages: []*api.LogMessage{{
				Time:  timestamppb.Now(),
				Level: api.LogLevel_ERROR,
				Text:  message,
			}},
		}
	}
	return state, nil
}

func (wh *WebHookHandler) postEnvelope(
	ctx context.Context, msgtype string, e envelope.Envelope, log zerolog.Logger,
) (*api.ActorProcessingState, error) {
	var state api.ActorProcessingState

	url := strings.ReplaceAll(wh.url, "{msgtype}", msgtype)

	payload, err := wh.envelopeMarshaler(ctx, e)
	if err != nil {
		return nil, err
	}

	log.Debug().
		Str("url", url).
		Str("envelope-id", e.ID().String()).
		Str("content-type", wh.contentType).
		Msg("posting envelope")

	req, err := http.NewRequest("POST", url, payload)
	if err != nil {
		return nil, fmt.Errorf("failed to forge the request: %w", err)
	}
	req = req.WithContext(ctx)
	req.Header.Set("Content-Type", wh.contentType)
	req.Header.Set("Accept", wh.contentType)
	for _, h := range wh.headers {
		req.Header.Set(h.Key, h.Value)
	}

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return nil, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] %s", err.Error()),
		}
	}
	defer resp.Body.Close()
	if httpCodeMatch(resp.StatusCode, wh.temporaryFailureHTTPCodes) {
		var (
			message = resp.Status
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message += string(buf[:n])
		} else {
			message += ("Could not read body: " + err.Error())
		}
		return nil, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] Temporary failure: %s", message),
		}
	}

	if httpCodeMatch(resp.StatusCode, wh.successHTTPCodes) {
		state.Status = api.ActorProcessingState_SUCCESS
	} else {
		state.Status = api.ActorProcessingState_ERROR
		var (
			message string
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message = string(buf[:n])
		} else {
			message = "Could not read body: " + err.Error()
		}
		message = fmt.Sprintf(`[http.webhook] Invalid http status. Expected one of %v, got %s.
The response body is:
%s`,
			wh.successHTTPCodes,
			resp.Status,
			message)
		state.Messages = []*api.LogMessage{
			{
				Time:  timestamppb.Now(),
				Level: api.LogLevel_ERROR,
				Text:  message,
			},
		}
	}
	return &state, nil
}

func (wh *WebHookHandler) postContent(
	ctx context.Context, msgtype string, _ string, payload io.Reader, log zerolog.Logger,
) (*api.ActorProcessingState, error) {
	var state api.ActorProcessingState

	url := strings.ReplaceAll(wh.url, "{msgtype}", msgtype)

	req, err := http.NewRequest("POST", url, payload)
	if err != nil {
		return nil, fmt.Errorf("failed to forge the request: %w", err)
	}
	req = req.WithContext(ctx)
	req.Header.Set("Content-Type", wh.contentType)
	req.Header.Set("Accept", wh.contentType)
	for _, h := range wh.headers {
		req.Header.Set(h.Key, h.Value)
	}

	log.Debug().
		Str("url", url).
		Str("msgtype", msgtype).
		Str("content-type", wh.contentType).
		Msg("posting content")
	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return nil, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] %s", err.Error()),
		}
	}
	defer resp.Body.Close()
	if httpCodeMatch(resp.StatusCode, wh.temporaryFailureHTTPCodes) {
		var (
			message = resp.Status
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message += string(buf[:n])
		} else {
			message += ("Could not read body: " + err.Error())
		}
		return nil, &nrpc.Error{
			Type:    nrpc.Error_SERVERTOOBUSY,
			Message: fmt.Sprintf("[http.webhook] Temporary failure: %s", message),
		}
	}

	if httpCodeMatch(resp.StatusCode, wh.successHTTPCodes) {
		state.Status = api.ActorProcessingState_SUCCESS
	} else {
		state.Status = api.ActorProcessingState_ERROR
		var (
			message string
			buf     = make([]byte, 1024*256)
		)
		if n, err := resp.Body.Read(buf); err == nil {
			message = string(buf[:n])
		} else {
			message = "Could not read body: " + err.Error()
		}
		message = fmt.Sprintf(`[http.webhook] Invalid http status. Expected one of %v, got %s.
The response body is:
%s`,
			wh.successHTTPCodes,
			resp.Status,
			message)
		state.Messages = []*api.LogMessage{
			{
				Time:  timestamppb.Now(),
				Level: api.LogLevel_ERROR,
				Text:  message,
			},
		}
	}
	return &state, nil
}

// Process a request
func (wh *WebHookHandler) Process(apc *xbus.ActorProcessingContext) error {
	var (
		state    *api.ActorProcessingState
		stateErr error
	)
	if apc.Log.GetLevel() <= zerolog.DebugLevel {
		inputs := make([]string, len(apc.Request.GetInputs()))
		for i, input := range apc.Request.GetInputs() {
			inputs[i] = input.GetName() + "/" + input.Envelope.GetIDAsUUID().String()
		}
		apc.Log.Debug().
			Str("payloadtype", string(wh.payloadType)).
			Strs("inputs", inputs).
			Msg("handling request")
	}
	switch wh.payloadType {
	case WebHookPayloadFullRequest:
		for _, input := range apc.Request.Inputs {
			e, err := apc.GetEnvelope(input.Name)
			if err != nil {
				return err
			}
			complete, err := envelope.CompleteWithContext(apc.Context, e)
			if err != nil {
				return err
			}

			input.Envelope = complete
			input.Position = nil
		}
		state, stateErr = wh.postRequest(apc.Context, "", apc.Request, apc.Log.Logger)
		if stateErr != nil {
			return stateErr
		}
	case WebHookPayloadEnvelope:
		state = &api.ActorProcessingState{
			Status: api.ActorProcessingState_SUCCESS,
		}
		for _, input := range apc.Request.Inputs {
			e, err := apc.GetEnvelope(input.Name)
			if err != nil {
				return err
			}
			s, err := wh.postEnvelope(apc.Context, "", e, apc.Log.Logger)
			if err != nil {
				return err
			}
			if s.Status != api.ActorProcessingState_SUCCESS {
				state = s
				break
			}
			state = s
		}
	case WebHookPayloadMessageInRequest,
		WebHookPayloadMessageInEnvelope,
		WebHookPayloadMessageContent:
		for _, input := range apc.Request.Inputs {
			e, err := apc.GetEnvelope(input.Name)
			if err != nil {
				return err
			}
			firstMessage := true
			delay := time.Second
			if err := envelope.ForEachMessage(e, func(msg envelope.Message) error {
				defer func() {
					delay = wh.retryInitialDelay
					firstMessage = false
				}()
				firstAttempt := time.Now()
				for {
					switch wh.payloadType {
					case WebHookPayloadMessageInRequest:
						request, err := requestWithSingleMessage(
							apc.Context, apc.Request, input.Name, msg)
						if err != nil {
							stateErr = err
						} else {
							state, stateErr = wh.postRequest(apc.Context, msg.MsgType(), request, apc.Log.Logger)
						}
					case WebHookPayloadMessageInEnvelope:
						env, err := envelopeWithSingleMessage(apc.Context, e, msg)
						if err != nil {
							stateErr = err
						} else {
							state, stateErr = wh.postEnvelope(apc.Context, msg.MsgType(), env, apc.Log.Logger)
						}
					case WebHookPayloadMessageContent:
						state, stateErr = wh.postContent(
							apc.Context, msg.MsgType(), wh.contentType, msg.Reader(), apc.Log.Logger,
						)
					}
					if stateErr != nil {
						var nrpcErr *nrpc.Error
						if errors.As(stateErr, &nrpcErr) && nrpcErr.Type == nrpc.Error_SERVERTOOBUSY {
							if firstMessage {
								log.Warn().
									Msg("server unvailable error on first try: returning SERVERTOOBUSY")
								return stateErr
							}
							if attemptSince := time.Since(firstAttempt); attemptSince > wh.retryMaxDelay {
								log.Warn().
									Dur("delay", attemptSince).
									Msg("server unvailable error and waited for too long: returning SERVERTOOBUSY")
								return stateErr
							}
							log.Warn().
								Dur("delay", delay).
								Msg("server unvailable error. waiting before new attempt")
							select {
							case <-apc.Context.Done():
								return stateErr
							case <-time.After(delay):
								delay = min(
									delay*(time.Duration(100+wh.retryBackoff))/100,
									wh.retryMaxDelay)
								log.Info().Msg("new attempt")
								continue
							}
						}
						state = &api.ActorProcessingState{
							Status: api.ActorProcessingState_ERROR,
							Messages: []*api.LogMessage{{
								Time:  timestamppb.Now(),
								Level: api.LogLevel_ERROR,
								Text:  stateErr.Error(),
							}},
						}
					}
					if state.Status != api.ActorProcessingState_SUCCESS {
						return ErrStopIteration
					}

					return nil
				}
			}); err != nil && !errors.Is(err, ErrStopIteration) {
				return err
			}
		}
	}

	if state.Status == api.ActorProcessingState_PENDING {
		apc.Detach()
	} else {
		apc.End(state.Status, state.Messages)
	}
	return nil
}

func init() {
	xbus.RegisterActorService(WebHookServiceName, NewWebHookHandler)
}

func httpCodeMatch(code int, list []int) bool {
	for _, value := range list {
		if value == code {
			return true
		}
	}
	return false
}

func parseHTTPCodesList(s string) ([]int, error) {
	var (
		splitted = strings.Split(s, ",")
		codes    = make([]int, 0, len(splitted))
	)
	for _, token := range splitted {
		token := strings.TrimSpace(token)
		if token == "" {
			continue
		}
		code, err := strconv.Atoi(token)
		if err != nil {
			return nil, fmt.Errorf("could not parse '%s': %w", token, err)
		}
		codes = append(codes, code)
	}

	return codes, nil
}

func parseHeadersList(s string) ([]Header, error) {
	var (
		splitted = strings.Split(s, "\n")
		res      = make([]Header, 0, len(splitted))
	)
	for _, line := range splitted {
		line = strings.TrimSpace(line)
		if line == "" {
			continue
		}
		sp := strings.SplitN(line, ":", 2)
		if len(sp) < 2 {
			return nil, fmt.Errorf("invalid header line: %s", line)
		}

		res = append(res, Header{
			Key:   strings.TrimSpace(sp[0]),
			Value: strings.TrimSpace(sp[1]),
		})
	}
	return res, nil
}

func parsePercentage(s string, defaultValue int) (int, error) {
	if s == "" {
		return defaultValue, nil
	}
	isratio := s[len(s)-1] != '%'
	v, err := strconv.Atoi(s)
	if err != nil {
		return 0, err
	}
	if isratio {
		v *= 100
	}
	return v, nil
}

func parseDuration(s string, defaultValue time.Duration) (time.Duration, error) {
	if s == "" {
		return defaultValue, nil
	}
	return time.ParseDuration(s)
}

func requestWithSingleMessage(
	_ context.Context, r *api.ActorProcessRequest, inputName string, msg envelope.Message,
) (*api.ActorProcessRequest, error) {
	if r == nil {
		panic("nil request")
	}
	var input *api.ActorProcessRequest_Input
	for _, i := range r.Inputs {
		if i.Name == inputName {
			input = i
			break
		}
	}
	if input == nil {
		return nil, fmt.Errorf("%w: input", ErrNoSuchInput)
	}

	chunks, err := envelope.ReadAllChunks(msg.ChunkReader())
	if err != nil {
		return nil, err
	}

	request := api.ActorProcessRequest{
		Context: r.Context,
		Inputs: []*api.ActorProcessRequest_Input{{
			Name: input.Name,
			Envelope: &api.Envelope{
				ID:       input.Envelope.ID,
				EventIDs: api.UUIDSlice{msg.ID()}.ToBytes(),
				Events: []*api.Event{{
					ID:        msg.ID().ToBytes(),
					Type:      msg.MsgType(),
					Index:     1,
					ItemCount: uint64(len(chunks)),
					Checksum:  msg.Checksum().AsUint32(),
					Items:     chunks,
				}},
			},
		}},
	}

	return &request, nil
}

func envelopeWithSingleMessage(
	_ context.Context, e envelope.Envelope, msg envelope.Message,
) (envelope.Envelope, error) {
	if e == nil {
		panic("nil envelope")
	}
	return envelope.NewEnvelopeWithID(e.ID(), msg), nil
}
